@@ -1,5 +1,32 @@
+#![allow(unstable_name_collisions)]
+
 //! This library provides a stable polyfill for Rust's [Strict Provenance] experiment.
 //!
+//! Mapping to STD APIs:
+//!
+//! ```rust ,ignore
+//! // core::ptr (sptr)
+//! pub fn invalid<T>(addr: usize) -> *const T;
+//! pub fn invalid_mut<T>(addr: usize) -> *mut T;
+//!
+//! // core::pointer (sptr::Strict)
+//! pub fn addr(self) -> usize;
+//! pub fn with_addr(self, addr: usize) -> Self;
+//! pub fn map_addr(self, f: impl FnOnce(usize) -> usize) -> Self;
+//!
+//! // NON-STANDARD EXTENSIONS (feature = uptr)
+//! sptr::uptr
+//! sptr::iptr
+//!
+//! // NON-STANDARD EXTENSIONS (feature = opaque_fn)
+//! sptr::OpaqueFn
+//!
+//! // DEPRECATED BY THIS MODEL in core::pointer (sptr::Strict)
+//! // (disable with `default-features = false`)
+//! pub fn to_bits(self) -> usize;
+//! pub fn from_bits(usize) -> Self;
+//! ```
+//!
 //! Swapping between the two should be as simple as switching between `sptr::` and `ptr::`
 //! for static functions. For methods, you must import `sptr::Strict` into your module for
 //! the extension trait's methods to overlay std. The compiler will (understandably)
@@ -19,25 +46,7 @@
 //! in your Cargo.toml.
 //!
 //! Rust is the canonical source of definitions for these APIs and semantics, but the docs
-//! here will vaguely try to mirror the docs checked into Rust. The current APIs are:
-//!
-//!
-//! ```rust ,ignore
-//! core::ptr (sptr)
-//! pub fn invalid<T>(addr: usize) -> *const T;
-//! pub fn invalid_mut<T>(addr: usize) -> *mut T;
-//!
-//! // core::pointer (sptr::Strict)
-//! pub fn addr(self) -> usize;
-//! pub fn with_addr(self, addr: usize) -> Self;
-//! pub fn map_addr(self, f: impl FnOnce(usize) -> usize) -> Self;
-//!
-//! // DEPRECATED BY THIS MODEL in core::pointer (sptr::Strict)
-//! // (disable with `default-features = false`)
-//! pub fn to_bits(self) -> usize;
-//! pub fn from_bits(usize) -> Self;
-//! ```
-//!
+//! here will vaguely try to mirror the docs checked into Rust.
 //!
 //! The following explanation of the model should also appear at the top of `core::ptr`:
 //!
@@ -356,7 +365,12 @@ pub const fn invalid_mut<T>(addr: usize) -> *mut T {
     addr as *mut T
 }
 
-pub trait Strict<T> {
+mod private {
+    pub trait Sealed {}
+}
+
+pub trait Strict: private::Sealed {
+    type Pointee;
     /// Gets the "address" portion of the pointer.
     ///
     /// This is equivalent to `self as usize`, which semantically discards
@@ -374,7 +388,7 @@ pub trait Strict<T> {
     #[must_use]
     fn addr(self) -> usize
     where
-        T: Sized;
+        Self::Pointee: Sized;
 
     /// Creates a new pointer with the given address.
     ///
@@ -393,7 +407,7 @@ pub trait Strict<T> {
     #[must_use]
     fn with_addr(self, addr: usize) -> Self
     where
-        T: Sized;
+        Self::Pointee: Sized;
 
     /// Creates a new pointer by mapping `self`'s address to a new one.
     ///
@@ -404,24 +418,29 @@ pub trait Strict<T> {
     #[must_use]
     fn map_addr(self, f: impl FnOnce(usize) -> usize) -> Self
     where
-        T: Sized;
+        Self::Pointee: Sized;
 
     #[must_use]
     #[deprecated = "to_bits is incompatible with strict_provenance"]
     #[cfg(feature = "deprecate_problems")]
     fn to_bits(self) -> usize
     where
-        T: Sized;
+        Self::Pointee: Sized;
 
     #[must_use]
     #[deprecated = "from_bits is incompatible with strict_provenance"]
     #[cfg(feature = "deprecate_problems")]
     fn from_bits(bits: usize) -> Self
     where
-        T: Sized;
+        Self::Pointee: Sized;
 }
 
-impl<T> Strict<T> for *mut T {
+impl<T> private::Sealed for *mut T {}
+impl<T> private::Sealed for *const T {}
+
+impl<T> Strict for *mut T {
+    type Pointee = T;
+
     #[must_use]
     #[inline]
     fn addr(self) -> usize
@@ -481,7 +500,9 @@ impl<T> Strict<T> for *mut T {
     }
 }
 
-impl<T> Strict<T> for *const T {
+impl<T> Strict for *const T {
+    type Pointee = T;
+
     #[must_use]
     #[inline]
     fn addr(self) -> usize
@@ -554,3 +575,13 @@ mod test {
         assert_eq!(val, 0);
     }
 }
+
+#[cfg(feature = "uptr")]
+pub mod int;
+#[cfg(feature = "uptr")]
+pub use int::*;
+
+#[cfg(feature = "opaque_fn")]
+pub mod func;
+#[cfg(feature = "opaque_fn")]
+pub use func::*;
